/**********************************************************************************************************************/
/*                                                VECTORS & MATRICES CLASSES                                          */
/* 17/09/97, M : creation of the file                                                                                 */
/* 25/12/97, M : some functions added from BMATH                                                                      */
/* 15/01/98, M : inlining bugfixed                                                                                    */
/* 25/11/99, M : rewriting for operators speed optims                                                                 */
/**********************************************************************************************************************/

/************************************************ INCLUDE *************************************************************/

#include  "stdafx.h"
#include  "CVectMat.inl"

#include  <math.h>                                                    // cos,sin,fabs

/**********************************************************************************************************************/
/*                                                3*3 MATRIX                                                          */
/**********************************************************************************************************************/

CMat3x3 CMat3x3::m3Inverse() const
  {                                                                   // Gauss-Jordan elimination with partial pivoting
  CMat3x3 m3A(*this);                                                 // as A evolves from the original  matrix to identity
  CMat3x3 m3B(m3Identity3D());                                        // ...B evolves from identity to inverse(A)
  BYTE    bI,bJ,bI1;

  // loop over cols of A from left to right, eliminating above & below diagonals

  for(bJ = 0; bJ < 3; bJ++)                                           // find largest pivot in column j among rows j..2
    { 
    bI1 = bJ;                                                         // row with largest pivot candidate
    for(bI = BYTE(bJ+1); bI < 3; bI++) if(fabs(m3A.m_v3V[bI].m_fV[bJ]) > fabs(m3A.m_v3V[bI1].m_fV[bJ])) bI1 = bI;

    // swap rows i1 & j in A & B to put pivot on diagonal
  
    vSwap(m3A.m_v3V[bI1], m3A.m_v3V[bJ]);
    vSwap(m3B.m_v3V[bI1], m3B.m_v3V[bJ]);

    // scale row j to have a unit diagonal
    
    FLOAT fK = m3A.m_v3V[bJ].m_fV[bJ];
    FLOAT fKinv = 1.0f/fK;
    m3B.m_v3V[bJ] *= fKinv;
    m3A.m_v3V[bJ] *= fKinv;

    // eliminate off-diagonal elements in col j of A, doing identical ops to B
    
    for(bI = 0; bI < 3; bI++) if(bI != bJ)
      { 
      m3B.m_v3V[bI] -= m3A.m_v3V[bI].m_fV[bJ]*m3B.m_v3V[bJ];
      m3A.m_v3V[bI] -= m3A.m_v3V[bI].m_fV[bJ]*m3A.m_v3V[bJ]; 
      }
    }

  return m3B;
  }

/**********************************************************************************************************************/
/*                                                4*4 MATRIX                                                          */
/**********************************************************************************************************************/

CMat4x4 CMat4x4::m4Inverse() const
  {                                                                   // Gauss-Jordan elimination with partial pivoting
  CMat4x4 m4A(*this);                                                 // as A evolves from the original  matrix to identity
  CMat4x4 m4B(m4Identity4D());                                        // ...B evolves from identity to inverse(A)
  BYTE    bI,bJ,bI1;

  // loop over cols of A from left to right, eliminating above & below diagonals

  for(bJ = 0; bJ < 4; bJ++)                                           // find largest pivot in column j among rows j..3
    { 
    bI1 = bJ;                                                         // row with largest pivot candidate
    for(bI = BYTE(bJ+1); bI < 4; bI++) if(fabs(m4A.m_v4V[bI].m_fV[bJ]) > fabs(m4A.m_v4V[bI1].m_fV[bJ])) bI1 = bI;

    // swap rows i1 & j in A & B to put pivot on diagonal
  
    vSwap(m4A.m_v4V[bI1], m4A.m_v4V[bJ]);
    vSwap(m4B.m_v4V[bI1], m4B.m_v4V[bJ]);

    // scale row j to have a unit diagonal
    
    FLOAT fK = m4A.m_v4V[bJ].m_fV[bJ];
    FLOAT fKinv = 1.0f/fK;
    m4B.m_v4V[bJ] *= fKinv;
    m4A.m_v4V[bJ] *= fKinv;

    // eliminate off-diagonal elements in col j of A, doing identical ops to B
    
    for(bI = 0; bI < 4; bI++) if(bI != bJ)
      { 
      m4B.m_v4V[bI] -= m4A.m_v4V[bI].m_fV[bJ]*m4B.m_v4V[bJ];
      m4A.m_v4V[bI] -= m4A.m_v4V[bI].m_fV[bJ]*m4A.m_v4V[bJ]; 
      }
    }

    return m4B;
  }

/**********************************************************************************************************************/
/*                                                TRANSFORMATIONS                                                     */
/**********************************************************************************************************************/

/************************************************ basic transformations ***********************************************/
// translation, rotation, scale, perspective
// 18/09/97, M
/**********************************************************************************************************************/

CMat3x3 m3Translation2D(const CVect2D& v2V)                           // 2D translation (3*3 matrix)
  { return CMat3x3(CVect3D(1.f, 0.f, v2V[_X_]),
                   CVect3D(0.f, 1.f, v2V[_Y_]),
                   CVect3D(0.f, 0.f, 1.f)); }

CMat4x4 m4Translation3D(const CVect3D& m_v3V)                         // 3D translation (4*4 matrix)
  { return CMat4x4(CVect4D(1.f, 0.f, 0.f, m_v3V[_X_]),
                   CVect4D(0.f, 1.f, 0.f, m_v3V[_Y_]),
                   CVect4D(0.f, 0.f, 1.f, m_v3V[_Z_]),
                   CVect4D(0.f, 0.f, 0.f, 1.f)); }

CMat3x3 m3Rotation2D(const CVect2D& v2Center,const FLOAT fAngleRad)
  {                                                                   // 2D rotation around a point (3*3 matrix)
  FLOAT fCos = (FLOAT)cos(fAngleRad);
  FLOAT fSin = (FLOAT)sin(fAngleRad);
  FLOAT fT   = (FLOAT)1.f-fCos;

  return CMat3x3(CVect3D(fCos, -fSin, v2Center[_X_]*fT+v2Center[_Y_]*fSin),
                 CVect3D(fSin,  fCos, v2Center[_Y_]*fT-v2Center[_X_]*fSin),
                 CVect3D(0.f,   0.f,  1.f)); 
  }

CMat4x4 m4Rotation3D(const CVect3D& v3Axis,const FLOAT fAngleRad)       
  {                                                                   // 3D rotation around an axis (4*4 matrix)
  FLOAT fCos = (FLOAT)cos(fAngleRad);
  FLOAT fSin = (FLOAT)sin(fAngleRad);
  FLOAT fT   = (FLOAT)1.f-fCos;
  FLOAT fSX  = v3Axis[_X_]*fSin, fSY = v3Axis[_Y_]*fSin, fSZ = v3Axis[_Z_]*fSin;
  FLOAT fTX  = v3Axis[_X_]*fT,   fTY = v3Axis[_Y_]*fT,   fTZ = v3Axis[_Z_]*fT;

  const_cast<CVect3D&>(v3Axis).boNormalize();

  return CMat4x4(CVect4D(fTX*v3Axis[_X_]+fCos, fTX*v3Axis[_Y_]-fSZ,  fTX*v3Axis[_Z_]+fSY,  0.f),
                 CVect4D(fTX*v3Axis[_Y_]+fSZ,  fTY*v3Axis[_Y_]+fCos, fTY*v3Axis[_Z_]-fSX,  0.f),
                 CVect4D(fTX*v3Axis[_Z_]-fSY,  fTY*v3Axis[_Z_]+fSX,  fTZ*v3Axis[_Z_]+fCos, 0.f),
                 CVect4D(       0.f,                  0.f,                  0.f,           1.f)); 
  }

CMat3x3 m3Scale2D(const CVect2D& v2Scale)                             // 2D scale (3*3 matrix)
  { return CMat3x3(CVect3D(v2Scale[_X_], 0.f, 0.f),
                   CVect3D(0.f, v2Scale[_Y_], 0.f),
                   CVect3D(0.f, 0.f, 1.f)); }

CMat4x4 m4Scale3D(const CVect3D& v3Scale)                             // 3D scale (4*4 matrix)
  { return CMat4x4(CVect4D(v3Scale[_X_], 0.f, 0.f, 0.f),
                   CVect4D(0.f, v3Scale[_Y_], 0.f, 0.f),
                   CVect4D(0.f, 0.f, v3Scale[_Z_], 0.f),
                   CVect4D(0.f, 0.f, 0.f, 1.f)); }

CMat4x4 m4Perspective3D(const FLOAT fD)                               // 3D perspective (4*4 matrix)
  {
  return CMat4x4(CVect4D(1.f, 0.f, 0.f, 0.f),
                 CVect4D(0.f, 1.f, 0.f, 0.f),
                 CVect4D(0.f, 0.f, 1.f, 0.f),
                 CVect4D(0.f, 0.f, (FLOAT)1.f/fD, 0.f)); 
  }

/**********************************************************************************************************************/
/*                                                CALCULATIONS                                                        */
/**********************************************************************************************************************/

/************************************************ vMultHomoMatrix *****************************************************/
// multiply two homogeneous matrices (last line = 0;0;0;1)
// 25/12/97, M
// in : result,M1,M2
// out: result=M1*M2
/**********************************************************************************************************************/

void vMultHomoMatrix(CMat4x4* lpm4M,const CMat4x4& m4M1,const CMat4x4& m4M2)
  {
  (*lpm4M)[0][0] = m4M1[0][0]*m4M2[0][0] + m4M1[0][1]*m4M2[1][0]
                 + m4M1[0][2]*m4M2[2][0];
  (*lpm4M)[0][1] = m4M1[0][0]*m4M2[0][1] + m4M1[0][1]*m4M2[1][1]
                 + m4M1[0][2]*m4M2[2][1];
  (*lpm4M)[0][2] = m4M1[0][0]*m4M2[0][2] + m4M1[0][1]*m4M2[1][2]
                 + m4M1[0][2]*m4M2[2][2];
  (*lpm4M)[0][3] = m4M1[0][0]*m4M2[0][3] + m4M1[0][1]*m4M2[1][3]
                 + m4M1[0][2]*m4M2[2][3] + m4M1[0][3];

  (*lpm4M)[1][0] = m4M1[1][0]*m4M2[0][0] + m4M1[1][1]*m4M2[1][0]
                 + m4M1[1][2]*m4M2[2][0];
  (*lpm4M)[1][1] = m4M1[1][0]*m4M2[0][1] + m4M1[1][1]*m4M2[1][1]
                 + m4M1[1][2]*m4M2[2][1];
  (*lpm4M)[1][2] = m4M1[1][0]*m4M2[0][2] + m4M1[1][1]*m4M2[1][2]
                 + m4M1[1][2]*m4M2[2][2];
  (*lpm4M)[1][3] = m4M1[1][0]*m4M2[0][3] + m4M1[1][1]*m4M2[1][3]
                 + m4M1[1][2]*m4M2[2][3] + m4M1[1][3];

  (*lpm4M)[2][0] = m4M1[2][0]*m4M2[0][0] + m4M1[2][1]*m4M2[1][0]
                 + m4M1[2][2]*m4M2[2][0];
  (*lpm4M)[2][1] = m4M1[2][0]*m4M2[0][1] + m4M1[2][1]*m4M2[1][1]
                 + m4M1[2][2]*m4M2[2][1];
  (*lpm4M)[2][2] = m4M1[2][0]*m4M2[0][2] + m4M1[2][1]*m4M2[1][2]
                 + m4M1[2][2]*m4M2[2][2];
  (*lpm4M)[2][3] = m4M1[2][0]*m4M2[0][3] + m4M1[2][1]*m4M2[1][3]
                 + m4M1[2][2]*m4M2[2][3] + m4M1[2][3];

  (*lpm4M)[3][0] = (*lpm4M)[3][1] = (*lpm4M)[3][2] = 0.f;
  (*lpm4M)[3][3] = 1.f;
  }

/************************************************ vTransposeHomoMatrix ************************************************/
// transpose a 4*4 tranformation (transpose rotation,change translation sign)
// 25/12/97, M
// in : result,M
// out: result=Mt
// rem: doesn't work if there are scale factors in the matrix (the scale is not reversed)
/**********************************************************************************************************************/

void vTransposeHomoMatrix(CMat4x4* lpm4Mt,const CMat4x4& m4M)
  {
  (*lpm4Mt)[0][0] = m4M[0][0];                                        // transpose rotation
  (*lpm4Mt)[0][1] = m4M[1][0];
  (*lpm4Mt)[0][2] = m4M[2][0];
  (*lpm4Mt)[1][0] = m4M[0][1];
  (*lpm4Mt)[1][1] = m4M[1][1];
  (*lpm4Mt)[1][2] = m4M[2][1];
  (*lpm4Mt)[2][0] = m4M[0][2];
  (*lpm4Mt)[2][1] = m4M[1][2];
  (*lpm4Mt)[2][2] = m4M[2][2];

  (*lpm4Mt)[0][3] =-m4M[0][3];                                        // reverse translation
  (*lpm4Mt)[1][3] =-m4M[1][3];
  (*lpm4Mt)[2][3] =-m4M[2][3];

  (*lpm4Mt)[3][0] = (*lpm4Mt)[3][1] = (*lpm4Mt)[3][2] = 0.f;          // last line
  (*lpm4Mt)[3][3] = 1.f;
  }

/************************************************ vRotate3DVector4x4 **************************************************/
// use the rotation part of a 4*4 matrix to rotate a 3*3 vector
// 25/12/97, M
// in : result,Matrix, Vector
// out: result=rotated Vector
/**********************************************************************************************************************/

void vRotate3DVector4x4(CVect3D* lpv3Vp,const CMat4x4& m4M,const CVect3D& v3V)
  {
  (*lpv3Vp)[_X_] = m4M[0][0]*v3V[_X_] + m4M[0][1]*v3V[_Y_] + m4M[0][2]*v3V[_Z_];
  (*lpv3Vp)[_Y_] = m4M[1][0]*v3V[_X_] + m4M[1][1]*v3V[_Y_] + m4M[1][2]*v3V[_Z_];
  (*lpv3Vp)[_Z_] = m4M[2][0]*v3V[_X_] + m4M[2][1]*v3V[_Y_] + m4M[2][2]*v3V[_Z_];
  }

/************************************************ vCalcRotMatrix ******************************************************/
// calculate 4x4 rotation matrix from 3 angles (Bryant)
// 25/12/97, M
// in : result,angles
// out:
/**********************************************************************************************************************/

void vCalcRotMatrix(CMat4x4* lpm4M,const CVect3D& v3Angles)
  {
  FLOAT c0,s0,c1,s1,c2,s2;
  c0 = (FLOAT)cos(v3Angles[_X_]); s0 = (FLOAT)sin(v3Angles[_X_]);
  c1 = (FLOAT)cos(v3Angles[_Y_]); s1 = (FLOAT)sin(v3Angles[_Y_]);
  c2 = (FLOAT)cos(v3Angles[_Z_]); s2 = (FLOAT)sin(v3Angles[_Z_]);
  (*lpm4M)[2][0] = -s1;

  FLOAT c0c1 = c0*c1;
  FLOAT s0c1 = s0*c1;
  (*lpm4M)[2][2] = c0c1;
  (*lpm4M)[2][1] = s0c1;

  FLOAT c1c2 = c1*c2;
  FLOAT c1s2 = c1*s2;
  (*lpm4M)[0][0] = c1c2;
  (*lpm4M)[1][0] = c1s2;

  FLOAT c0s2 = c0*s2;
  FLOAT c0c2 = c0*c2;
  FLOAT s0s2 = s0*s2;
  FLOAT s0c2 = s0*c2;
  (*lpm4M)[0][1] =-c0s2+s0c2*s1;
  (*lpm4M)[1][1] = c0c2+s0s2*s1;
  (*lpm4M)[0][2] = s0s2+c0c2*s1;
  (*lpm4M)[1][2] =-s0c2+c0s2*s1;
  }

/************************************************ vCalcRotMatrix ******************************************************/
// calculate 3x3 rotation matrix from 3 angles (Bryant)
// 01/10/99
// in : result,angles
// out:
/**********************************************************************************************************************/

void vCalcRotMatrix(CMat3x3* lpm3M,const CVect3D& v3Angles)
  {
  FLOAT c0,s0,c1,s1,c2,s2;
  c0 = (FLOAT)cos(v3Angles[_X_]); s0 = (FLOAT)sin(v3Angles[_X_]);
  c1 = (FLOAT)cos(v3Angles[_Y_]); s1 = (FLOAT)sin(v3Angles[_Y_]);
  c2 = (FLOAT)cos(v3Angles[_Z_]); s2 = (FLOAT)sin(v3Angles[_Z_]);
  (*lpm3M)[2][0] = -s1;

  FLOAT c0c1 = c0*c1;
  FLOAT s0c1 = s0*c1;
  (*lpm3M)[2][2] = c0c1;
  (*lpm3M)[2][1] = s0c1;

  FLOAT c1c2 = c1*c2;
  FLOAT c1s2 = c1*s2;
  (*lpm3M)[0][0] = c1c2;
  (*lpm3M)[1][0] = c1s2;

  FLOAT c0s2 = c0*s2;
  FLOAT c0c2 = c0*c2;
  FLOAT s0s2 = s0*s2;
  FLOAT s0c2 = s0*c2;
  (*lpm3M)[0][1] =-c0s2+s0c2*s1;
  (*lpm3M)[1][1] = c0c2+s0s2*s1;
  (*lpm3M)[0][2] = s0s2+c0c2*s1;
  (*lpm3M)[1][2] =-s0c2+c0s2*s1;
}

/**********************************************************************************************************************/
/*                                                2D                                                                  */
/**********************************************************************************************************************/

void vAdd(CVect2D* lpv2V, const CVect2D& v2V1)                        // V+=V1
  {
  (*lpv2V)[0] += v2V1[0];
  (*lpv2V)[1] += v2V1[1];
  }

void vSub(CVect2D* lpv2V, const CVect2D& v2V1)                        // V-=V1
  {
  (*lpv2V)[0] -= v2V1[0];
  (*lpv2V)[1] -= v2V1[1];
  }

void vMul(CVect2D* lpv2V, const FLOAT fK)                             // V*=k
  {
  (*lpv2V)[0] *= fK;
  (*lpv2V)[1] *= fK;
  }

void vMul(CVect2D* lpv2V, const CVect2D& v2V1)                        // V*=V1 (term by term)
  {
  (*lpv2V)[0] *= v2V1[0];
  (*lpv2V)[1] *= v2V1[1];
  }

void vDiv(CVect2D* lpv2V, const FLOAT fK)                             // V/=k
  {
  FLOAT fKinv = 1.f/fK;
  (*lpv2V)[0] *= fKinv;
  (*lpv2V)[1] *= fKinv;
  }

void vDiv(CVect2D* lpv2V, const CVect2D& v2V1)                        // V/=V1 (term by term)
  {
  (*lpv2V)[0] /= v2V1[0];
  (*lpv2V)[1] /= v2V1[1];
  }

void vNeg(CVect2D* lpv2V, const CVect2D& v2V1)                        // V=-V1
  {
  (*lpv2V)[0] = -v2V1[0];
  (*lpv2V)[1] = -v2V1[1];
  }

void vMul(CVect2D* lpv2V, const CVect2D& v2V1,const FLOAT fK)
  {                                                                   // V=V1*k
  (*lpv2V)[0] = v2V1[0]*fK;
  (*lpv2V)[1] = v2V1[1]*fK;
  }

void vDiv(CVect2D* lpv2V, const CVect2D& v2V1,const FLOAT fK)
  {                                                                   // V=V1/k
  FLOAT fKinv = 1.f/fK;
  (*lpv2V)[0] = v2V1[0]*fKinv;
  (*lpv2V)[1] = v2V1[1]*fKinv;
  }

void vSub(CVect2D* lpv2V, const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // V=V1-V2
  (*lpv2V)[0] = v2V1[0] - v2V2[0];
  (*lpv2V)[1] = v2V1[1] - v2V2[1];
  }

void vAdd(CVect2D* lpv2V, const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // V=V1+V2
  (*lpv2V)[0] = v2V1[0] + v2V2[0];
  (*lpv2V)[1] = v2V1[1] + v2V2[1];
  }

FLOAT fDotProd  (const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // f=V1*V2
  return(v2V1[0]*v2V2[0] + v2V1[1]*v2V2[1]);
  }

FLOAT fCrossProd(const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // f=V1^V2
  return(v2V1[0]*v2V2[1] - v2V1[1]*v2V2[0]);
  }

void vMul(CVect2D* lpv2V, const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // V=V1*V2 (term by term)
  (*lpv2V)[0] = v2V1[0] * v2V2[0];
  (*lpv2V)[1] = v2V1[1] * v2V2[1];
  }

void vDiv(CVect2D* lpv2V, const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // V=V1/V2 (term by term)
  (*lpv2V)[0] = v2V1[0] / v2V2[0];
  (*lpv2V)[1] = v2V1[1] / v2V2[1];
  }

/**********************************************************************************************************************/
/*                                                3D                                                                  */
/**********************************************************************************************************************/

void vAdd(CVect3D* lpv3V, const CVect3D& v3V1)                        // V+=V1
  {
  (*lpv3V)[0] += v3V1[0];
  (*lpv3V)[1] += v3V1[1];
  (*lpv3V)[2] += v3V1[2];
  }
          
void vSub(CVect3D* lpv3V, const CVect3D& v3V1)                        // V-=V1
  {
  (*lpv3V)[0] -= v3V1[0];
  (*lpv3V)[1] -= v3V1[1];
  (*lpv3V)[2] -= v3V1[2];
  }          

void vMul(CVect3D* lpv3V, const FLOAT fK)                             // V*=k
  {
  (*lpv3V)[0] *= fK;
  (*lpv3V)[1] *= fK;
  (*lpv3V)[2] *= fK;          
  }

void vMul(CVect3D* lpv3V, const CVect3D& v3V1)                        // V*=V1 (term by term)
  {
  (*lpv3V)[0] *= v3V1[0];
  (*lpv3V)[1] *= v3V1[1];          
  (*lpv3V)[2] *= v3V1[2];
  }

void vDiv(CVect3D* lpv3V, const FLOAT fK)                             // V/=k
  {
  FLOAT fKinv = 1.f/fK;
  (*lpv3V)[0] *= fKinv;          
  (*lpv3V)[1] *= fKinv;
  (*lpv3V)[2] *= fKinv;
  }

void vDiv(CVect3D* lpv3V, const CVect3D& v3V1)                        // V/=V1 (term by term)
  {
  (*lpv3V)[0] /= v3V1[0];
  (*lpv3V)[1] /= v3V1[1];
  (*lpv3V)[2] /= v3V1[2];
  }

void vNeg(CVect3D* lpv3V, const CVect3D& v3V1)                        // V=-V1
  {
  (*lpv3V)[0] = -v3V1[0];
  (*lpv3V)[1] = -v3V1[1];
  (*lpv3V)[2] = -v3V1[2];
  }

void vMul(CVect3D* lpv3V, const CVect3D& v3V1,const FLOAT fK)
  {                                                                   // V=V1*k
  (*lpv3V)[0] = v3V1[0]*fK;
  (*lpv3V)[1] = v3V1[1]*fK;
  (*lpv3V)[2] = v3V1[2]*fK;
  }

void vDiv(CVect3D* lpv3V, const CVect3D& v3V1,const FLOAT fK)
  {                                                                   // V=V1/k
  FLOAT fKinv = 1.f/fK;
  (*lpv3V)[0] = v3V1[0]*fKinv;
  (*lpv3V)[1] = v3V1[1]*fKinv;
  (*lpv3V)[2] = v3V1[2]*fKinv;
  }

void vSub(CVect3D* lpv3V, const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1-V2
  (*lpv3V)[0] = v3V1[0] - v3V2[0];
  (*lpv3V)[1] = v3V1[1] - v3V2[1];
  (*lpv3V)[2] = v3V1[2] - v3V2[2];
  }

void vAdd(CVect3D* lpv3V, const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1+V2
  (*lpv3V)[0] = v3V1[0] + v3V2[0];
  (*lpv3V)[1] = v3V1[1] + v3V2[1];
  (*lpv3V)[2] = v3V1[2] + v3V2[2];
  }

FLOAT fDotProd(const CVect3D& v3V1,const CVect3D& v3V2)               // f=V1*V2
  {
  return(v3V1[0]*v3V2[0] + v3V1[1]*v3V2[1] + v3V1[2]*v3V2[2]);
  }

FLOAT fDotProd(const CVect4D& v4V1,const CVect3D& v3V2)               // f=V1*V2
  {
  return(v4V1[0]*v3V2[0] + v4V1[1]*v3V2[1] + v4V1[2]*v3V2[2] + v4V1[3]);
  }

void vCrossProd(CVect3D* lpv3V,const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1^V2
  (*lpv3V)[0] = v3V1[_Y_]*v3V2[_Z_] - v3V1[_Z_]*v3V2[_Y_];
  (*lpv3V)[1] = v3V1[_Z_]*v3V2[_X_] - v3V1[_X_]*v3V2[_Z_];
  (*lpv3V)[2] = v3V1[_X_]*v3V2[_Y_] - v3V1[_Y_]*v3V2[_X_];
  }

void vMul(CVect3D* lpv3V, const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1*V2 (term by term)
  (*lpv3V)[0] = v3V1[0] * v3V2[0];
  (*lpv3V)[1] = v3V1[1] * v3V2[1];
  (*lpv3V)[2] = v3V1[2] * v3V2[2];
  }

void vDiv(CVect3D* lpv3V, const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1/V2 (term by term)
  (*lpv3V)[0] = v3V1[0] / v3V2[0];
  (*lpv3V)[1] = v3V1[1] / v3V2[1];
  (*lpv3V)[2] = v3V1[2] / v3V2[2];
  }

void vMul(CVect3D* lpv3V, const CMat3x3& m3M ,const CVect3D& v3V1)    // V=M*V1
  {
  (*lpv3V)[0] = m3M[0][0]*v3V1[0] + m3M[0][1]*v3V1[1] + m3M[0][2]*v3V1[2];
  (*lpv3V)[1] = m3M[1][0]*v3V1[0] + m3M[1][1]*v3V1[1] + m3M[1][2]*v3V1[2];
  (*lpv3V)[2] = m3M[2][0]*v3V1[0] + m3M[2][1]*v3V1[1] + m3M[2][2]*v3V1[2];
  }

void vMul(CVect3D* lpv3V, const CVect3D& v3V1,const CMat3x3& m3M)     // V=V1*M
  {
  (*lpv3V)[0] = m3M[0][0]*v3V1[0] + m3M[1][0]*v3V1[1] + m3M[2][0]*v3V1[2];
  (*lpv3V)[1] = m3M[0][1]*v3V1[0] + m3M[1][1]*v3V1[1] + m3M[2][1]*v3V1[2];
  (*lpv3V)[2] = m3M[0][2]*v3V1[0] + m3M[1][2]*v3V1[1] + m3M[2][2]*v3V1[2];
  }

void vMul(CVect4D* lpv4V, const CMat4x4& m4M ,const CVect3D& v3V1)    // V=M*V1
  {
  (*lpv4V)[0] = m4M[0][0]*v3V1[0] + m4M[0][1]*v3V1[1] + m4M[0][2]*v3V1[2] + m4M[0][3];
  (*lpv4V)[1] = m4M[1][0]*v3V1[0] + m4M[1][1]*v3V1[1] + m4M[1][2]*v3V1[2] + m4M[1][3];
  (*lpv4V)[2] = m4M[2][0]*v3V1[0] + m4M[2][1]*v3V1[1] + m4M[2][2]*v3V1[2] + m4M[2][3];
  (*lpv4V)[3] = m4M[3][0]*v3V1[0] + m4M[3][1]*v3V1[1] + m4M[3][2]*v3V1[2] + m4M[3][3];
  }

void vMul(CVect4D* lpv4V, const CVect3D& v3V1,const CMat4x4& m4M)     // V=V1*M
  {
  (*lpv4V)[0] = m4M[0][0]*v3V1[0] + m4M[1][0]*v3V1[1] + m4M[2][0]*v3V1[2] + m4M[3][0];
  (*lpv4V)[1] = m4M[0][1]*v3V1[0] + m4M[1][1]*v3V1[1] + m4M[2][1]*v3V1[2] + m4M[3][1];
  (*lpv4V)[2] = m4M[0][2]*v3V1[0] + m4M[1][2]*v3V1[1] + m4M[2][2]*v3V1[2] + m4M[3][2];
  (*lpv4V)[3] = m4M[0][3]*v3V1[0] + m4M[1][3]*v3V1[1] + m4M[2][3]*v3V1[2] + m4M[3][3];
  }

/**********************************************************************************************************************/
/*                                                4D                                                                  */
/**********************************************************************************************************************/

void vAdd(CVect4D* lpv4V, const CVect4D& v4V1)                        // V+=V1
  {
  (*lpv4V)[0] += v4V1[0];
  (*lpv4V)[1] += v4V1[1];
  (*lpv4V)[2] += v4V1[2];
  (*lpv4V)[3] += v4V1[3];
  }

void vSub(CVect4D* lpv4V, const CVect4D& v4V1)                        // V-=V1
  {
  (*lpv4V)[0] -= v4V1[0];
  (*lpv4V)[1] -= v4V1[1];
  (*lpv4V)[2] -= v4V1[2];
  (*lpv4V)[3] -= v4V1[3];
  }

void vMul(CVect4D* lpv4V, const FLOAT fK)                             // V*=k
  {
  (*lpv4V)[0] *= fK;
  (*lpv4V)[1] *= fK;
  (*lpv4V)[2] *= fK;
  (*lpv4V)[3] *= fK;
  }

void vMul(CVect4D* lpv4V, const CVect4D& v4V1)                        // V*=V1 (term by term)
  {
  (*lpv4V)[0] *= v4V1[0];
  (*lpv4V)[1] *= v4V1[1];
  (*lpv4V)[2] *= v4V1[2];
  (*lpv4V)[3] *= v4V1[3];
  }

void vDiv(CVect4D* lpv4V, const FLOAT fK)                             // V/=k
  {
  FLOAT fKinv = 1.f/fK;
  (*lpv4V)[0] *= fKinv;
  (*lpv4V)[1] *= fKinv;
  (*lpv4V)[2] *= fKinv;
  (*lpv4V)[3] *= fKinv;
  }

void vDiv(CVect4D* lpv4V, const CVect4D& v4V1)                        // V/=V1 (term by term)
  {
  (*lpv4V)[0] /= v4V1[0];
  (*lpv4V)[1] /= v4V1[1];
  (*lpv4V)[2] /= v4V1[2];
  (*lpv4V)[3] /= v4V1[3];
  }

void vNeg(CVect4D* lpv4V, const CVect4D& v4V1)                        // V=-V1
  {
  (*lpv4V)[0] = -v4V1[0];
  (*lpv4V)[1] = -v4V1[1];
  (*lpv4V)[2] = -v4V1[2];
  (*lpv4V)[3] = -v4V1[3];
  }

void vMul(CVect4D* lpv4V, const CVect4D& v4V1,FLOAT fK)               // V=V1*k
  {
  (*lpv4V)[0] = v4V1[0]*fK;
  (*lpv4V)[1] = v4V1[1]*fK;
  (*lpv4V)[2] = v4V1[2]*fK;
  (*lpv4V)[3] = v4V1[3]*fK;
  }

void vDiv(CVect4D* lpv4V, const CVect4D& v4V1,const FLOAT fK)         // V=V1/k
  {
  FLOAT fKinv = 1.f/fK;
  (*lpv4V)[0] = v4V1[0]*fKinv;
  (*lpv4V)[1] = v4V1[1]*fKinv;
  (*lpv4V)[2] = v4V1[2]*fKinv;
  (*lpv4V)[3] = v4V1[3]*fKinv;
  }

void vSub(CVect4D* lpv4V, const CVect4D& v4V1,const CVect4D& v4V2)    // V=V1-V2
  {
  (*lpv4V)[0] = v4V1[0] - v4V2[0];
  (*lpv4V)[1] = v4V1[1] - v4V2[1];
  (*lpv4V)[2] = v4V1[2] - v4V2[2];
  (*lpv4V)[3] = v4V1[3] - v4V2[3];
  }

void vAdd(CVect4D* lpv4V, const CVect4D& v4V1,const CVect4D& v4V2)    // V=V1+V2
  {
  (*lpv4V)[0] = v4V1[0] + v4V2[0];
  (*lpv4V)[1] = v4V1[1] + v4V2[1];
  (*lpv4V)[2] = v4V1[2] + v4V2[2];
  (*lpv4V)[3] = v4V1[3] + v4V2[3];
  }

FLOAT fDotProd(const CVect4D& v4V1,const CVect4D& v4V2)               // f=V1*V2
  {
  return(v4V1[0]*v4V2[0] + v4V1[1]*v4V2[1] + v4V1[2]*v4V2[2] + v4V1[3]*v4V2[3]);
  }

void vMul(CVect4D* lpv4V, const CVect4D& v4V1,const CVect4D& v4V2)    // V=V1*V2 (term by term)
  {
  (*lpv4V)[0] = v4V1[0] * v4V2[0];
  (*lpv4V)[1] = v4V1[1] * v4V2[1];
  (*lpv4V)[2] = v4V1[2] * v4V2[2];
  (*lpv4V)[3] = v4V1[3] * v4V2[3];
  }

void vDiv(CVect4D* lpv4V, const CVect4D& v4V1,const CVect4D& v4V2)    // V=V1/V2 (term by term)
  {
  (*lpv4V)[0] = v4V1[0] / v4V2[0];
  (*lpv4V)[1] = v4V1[1] / v4V2[1];
  (*lpv4V)[2] = v4V1[2] / v4V2[2];
  (*lpv4V)[3] = v4V1[3] / v4V2[3];
  }

void vMul(CVect4D* lpv4V, const CMat4x4& m4M ,const CVect4D& v4V1)    // V=M*V1
  {
  (*lpv4V)[0] = m4M[0][0]*v4V1[0] + m4M[0][1]*v4V1[1] + m4M[0][2]*v4V1[2] + m4M[0][3]*v4V1[3];
  (*lpv4V)[1] = m4M[1][0]*v4V1[0] + m4M[1][1]*v4V1[1] + m4M[1][2]*v4V1[2] + m4M[1][3]*v4V1[3];
  (*lpv4V)[2] = m4M[2][0]*v4V1[0] + m4M[2][1]*v4V1[1] + m4M[2][2]*v4V1[2] + m4M[2][3]*v4V1[3];
  (*lpv4V)[3] = m4M[3][0]*v4V1[0] + m4M[3][1]*v4V1[1] + m4M[3][2]*v4V1[2] + m4M[3][3]*v4V1[3];
  }

void vMul(CVect4D* lpv4V, const CVect4D& v4V1,const CMat4x4& m4M)     // V=V1*M
  {
  (*lpv4V)[0] = m4M[0][0]*v4V1[0] + m4M[1][0]*v4V1[1] + m4M[2][0]*v4V1[2] + m4M[3][0]*v4V1[3];
  (*lpv4V)[1] = m4M[0][1]*v4V1[0] + m4M[1][1]*v4V1[1] + m4M[2][1]*v4V1[2] + m4M[3][1]*v4V1[3];
  (*lpv4V)[2] = m4M[0][2]*v4V1[0] + m4M[1][2]*v4V1[1] + m4M[2][2]*v4V1[2] + m4M[3][2]*v4V1[3];
  (*lpv4V)[3] = m4M[0][3]*v4V1[0] + m4M[1][3]*v4V1[1] + m4M[2][3]*v4V1[2] + m4M[3][3]*v4V1[3];
  }

/**********************************************************************************************************************/
/*                                                3x3                                                                 */
/**********************************************************************************************************************/

void vAdd(CMat3x3* lpm3M, const CMat3x3& m3M1)                        // M+=M1
  {
  (*lpm3M)[0][0] += m3M1[0][0];
  (*lpm3M)[0][1] += m3M1[0][1];
  (*lpm3M)[0][2] += m3M1[0][2];

  (*lpm3M)[1][0] += m3M1[1][0];
  (*lpm3M)[1][1] += m3M1[1][1];
  (*lpm3M)[1][2] += m3M1[1][2];

  (*lpm3M)[2][0] += m3M1[2][0];
  (*lpm3M)[2][1] += m3M1[2][1];
  (*lpm3M)[2][2] += m3M1[2][2];
  }

void vSub(CMat3x3* lpm3M, const CMat3x3& m3M1)                        // M-=M1
  {
  (*lpm3M)[0][0] -= m3M1[0][0];
  (*lpm3M)[0][1] -= m3M1[0][1];
  (*lpm3M)[0][2] -= m3M1[0][2];

  (*lpm3M)[1][0] -= m3M1[1][0];
  (*lpm3M)[1][1] -= m3M1[1][1];
  (*lpm3M)[1][2] -= m3M1[1][2];

  (*lpm3M)[2][0] -= m3M1[2][0];
  (*lpm3M)[2][1] -= m3M1[2][1];
  (*lpm3M)[2][2] -= m3M1[2][2];
  }

void vMul(CMat3x3* lpm3M, const FLOAT fK)                             // M*=k
  {
  (*lpm3M)[0][0] *= fK;
  (*lpm3M)[0][1] *= fK;
  (*lpm3M)[0][2] *= fK;

  (*lpm3M)[1][0] *= fK;
  (*lpm3M)[1][1] *= fK;
  (*lpm3M)[1][2] *= fK;

  (*lpm3M)[2][0] *= fK;
  (*lpm3M)[2][1] *= fK;
  (*lpm3M)[2][2] *= fK;
  }

void vMul(CMat3x3* lpm3M, const CMat3x3& m3M1)                        // M*=M1
  {
  *lpm3M = (*lpm3M)*m3M1;
  }

void vDiv(CMat3x3* lpm3M, const FLOAT fK)                             // M/=k
  {
  FLOAT fKinv = 1.f/fK;

  (*lpm3M)[0][0] *= fKinv;
  (*lpm3M)[0][1] *= fKinv;
  (*lpm3M)[0][2] *= fKinv;

  (*lpm3M)[1][0] *= fKinv;
  (*lpm3M)[1][1] *= fKinv;
  (*lpm3M)[1][2] *= fKinv;

  (*lpm3M)[2][0] *= fKinv;
  (*lpm3M)[2][1] *= fKinv;
  (*lpm3M)[2][2] *= fKinv;
  }

void vDiv(CMat3x3* lpm3M, const CMat3x3& m3M1)                        // M/=M1 (* inverse)
  {
  *lpm3M = (*lpm3M)*(m3M1.m3Inverse());
  }

void vNeg(CMat3x3* lpm3M, const CMat3x3& m3M1)                        // M=-M1
  {
  (*lpm3M)[0][0] = -m3M1[0][0];
  (*lpm3M)[0][1] = -m3M1[0][1];
  (*lpm3M)[0][2] = -m3M1[0][2];

  (*lpm3M)[1][0] = -m3M1[1][0];
  (*lpm3M)[1][1] = -m3M1[1][1];
  (*lpm3M)[1][2] = -m3M1[1][2];

  (*lpm3M)[2][0] = -m3M1[2][0];
  (*lpm3M)[2][1] = -m3M1[2][1];
  (*lpm3M)[2][2] = -m3M1[2][2];
  }

void vMul(CMat3x3* lpm3M, const CMat3x3& m3M1,const FLOAT fK)         // M=M1*k
  {
  (*lpm3M)[0][0] = m3M1[0][0]*fK;
  (*lpm3M)[0][1] = m3M1[0][1]*fK;
  (*lpm3M)[0][2] = m3M1[0][2]*fK;

  (*lpm3M)[1][0] = m3M1[1][0]*fK;
  (*lpm3M)[1][1] = m3M1[1][1]*fK;
  (*lpm3M)[1][2] = m3M1[1][2]*fK;

  (*lpm3M)[2][0] = m3M1[2][0]*fK;
  (*lpm3M)[2][1] = m3M1[2][1]*fK;
  (*lpm3M)[2][2] = m3M1[2][2]*fK;
  }

void vDiv(CMat3x3* lpm3M, const CMat3x3& m3M1,const FLOAT fK)         // M=M1/k
  {
  FLOAT fKinv = 1.f/fK;

  (*lpm3M)[0][0] = m3M1[0][0]*fKinv;
  (*lpm3M)[0][1] = m3M1[0][1]*fKinv;
  (*lpm3M)[0][2] = m3M1[0][2]*fKinv;

  (*lpm3M)[1][0] = m3M1[1][0]*fKinv;
  (*lpm3M)[1][1] = m3M1[1][1]*fKinv;
  (*lpm3M)[1][2] = m3M1[1][2]*fKinv;

  (*lpm3M)[2][0] = m3M1[2][0]*fKinv;
  (*lpm3M)[2][1] = m3M1[2][1]*fKinv;
  (*lpm3M)[2][2] = m3M1[2][2]*fKinv;
  }

void vSub(CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)    // M=M1-M2
  {
  (*lpm3M)[0][0] = m3M1[0][0] - m3M2[0][0];
  (*lpm3M)[0][1] = m3M1[0][1] - m3M2[0][1];
  (*lpm3M)[0][2] = m3M1[0][2] - m3M2[0][2];

  (*lpm3M)[1][0] = m3M1[1][0] - m3M2[1][0];
  (*lpm3M)[1][1] = m3M1[1][1] - m3M2[1][1];
  (*lpm3M)[1][2] = m3M1[1][2] - m3M2[1][2];

  (*lpm3M)[2][0] = m3M1[2][0] - m3M2[2][0];
  (*lpm3M)[2][1] = m3M1[2][1] - m3M2[2][1];
  (*lpm3M)[2][2] = m3M1[2][2] - m3M2[2][2];
  }

void vAdd(CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)    // M=M1+M2
  {
  (*lpm3M)[0][0] = m3M1[0][0] + m3M2[0][0];
  (*lpm3M)[0][1] = m3M1[0][1] + m3M2[0][1];
  (*lpm3M)[0][2] = m3M1[0][2] + m3M2[0][2];

  (*lpm3M)[1][0] = m3M1[1][0] + m3M2[1][0];
  (*lpm3M)[1][1] = m3M1[1][1] + m3M2[1][1];
  (*lpm3M)[1][2] = m3M1[1][2] + m3M2[1][2];

  (*lpm3M)[2][0] = m3M1[2][0] + m3M2[2][0];
  (*lpm3M)[2][1] = m3M1[2][1] + m3M2[2][1];
  (*lpm3M)[2][2] = m3M1[2][2] + m3M2[2][2];
  }

void vMul(CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)    // M=M1*M2
  {
  (*lpm3M)[0][0] = m3M1[0][0]*m3M2[0][0] + m3M1[0][1]*m3M2[1][0]
                 + m3M1[0][2]*m3M2[2][0];
  (*lpm3M)[0][1] = m3M1[0][0]*m3M2[0][1] + m3M1[0][1]*m3M2[1][1]
                 + m3M1[0][2]*m3M2[2][1];
  (*lpm3M)[0][2] = m3M1[0][0]*m3M2[0][2] + m3M1[0][1]*m3M2[1][2]
                 + m3M1[0][2]*m3M2[2][2];

  (*lpm3M)[1][0] = m3M1[1][0]*m3M2[0][0] + m3M1[1][1]*m3M2[1][0]
                 + m3M1[1][2]*m3M2[2][0];
  (*lpm3M)[1][1] = m3M1[1][0]*m3M2[0][1] + m3M1[1][1]*m3M2[1][1]
                 + m3M1[1][2]*m3M2[2][1];
  (*lpm3M)[1][2] = m3M1[1][0]*m3M2[0][2] + m3M1[1][1]*m3M2[1][2]
                 + m3M1[1][2]*m3M2[2][2];

  (*lpm3M)[2][0] = m3M1[2][0]*m3M2[0][0] + m3M1[2][1]*m3M2[1][0]
                 + m3M1[2][2]*m3M2[2][0];
  (*lpm3M)[2][1] = m3M1[2][0]*m3M2[0][1] + m3M1[2][1]*m3M2[1][1]
                 + m3M1[2][2]*m3M2[2][1];
  (*lpm3M)[2][2] = m3M1[2][0]*m3M2[0][2] + m3M1[2][1]*m3M2[1][2]
                 + m3M1[2][2]*m3M2[2][2];
  }

void vMulTerm(CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)
  {                                                                   // M=M1*M2 (term by term)
  (*lpm3M)[0][0] = m3M1[0][0] * m3M2[0][0];
  (*lpm3M)[0][1] = m3M1[0][1] * m3M2[0][1];
  (*lpm3M)[0][2] = m3M1[0][2] * m3M2[0][2];

  (*lpm3M)[1][0] = m3M1[1][0] * m3M2[1][0];
  (*lpm3M)[1][1] = m3M1[1][1] * m3M2[1][1];
  (*lpm3M)[1][2] = m3M1[1][2] * m3M2[1][2];

  (*lpm3M)[2][0] = m3M1[2][0] * m3M2[2][0];
  (*lpm3M)[2][1] = m3M1[2][1] * m3M2[2][1];
  (*lpm3M)[2][2] = m3M1[2][2] * m3M2[2][2];
  }

void vDiv(CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)    // M=M1/M2 (term by term)
  {
  (*lpm3M)[0][0] = m3M1[0][0] / m3M2[0][0];
  (*lpm3M)[0][1] = m3M1[0][1] / m3M2[0][1];
  (*lpm3M)[0][2] = m3M1[0][2] / m3M2[0][2];

  (*lpm3M)[1][0] = m3M1[1][0] / m3M2[1][0];
  (*lpm3M)[1][1] = m3M1[1][1] / m3M2[1][1];
  (*lpm3M)[1][2] = m3M1[1][2] / m3M2[1][2];

  (*lpm3M)[2][0] = m3M1[2][0] / m3M2[2][0];
  (*lpm3M)[2][1] = m3M1[2][1] / m3M2[2][1];
  (*lpm3M)[2][2] = m3M1[2][2] / m3M2[2][2];
  }

void vTranspose(CMat3x3* lpm3M, const CMat3x3& m3M1)                  // M=tM1
  {
  (*lpm3M)[0][0] = m3M1[0][0];
  (*lpm3M)[0][1] = m3M1[1][0];
  (*lpm3M)[0][2] = m3M1[2][0];

  (*lpm3M)[1][0] = m3M1[0][1];
  (*lpm3M)[1][1] = m3M1[1][1];
  (*lpm3M)[1][2] = m3M1[2][1];

  (*lpm3M)[2][0] = m3M1[0][2];
  (*lpm3M)[2][1] = m3M1[1][2];
  (*lpm3M)[2][2] = m3M1[2][2];
  }

void vMulMxtM(CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)
  {                                                                   // M=M1*tM2
  (*lpm3M)[0][0] = m3M1[0][0]*m3M2[0][0] + m3M1[0][1]*m3M2[0][1]
                 + m3M1[0][2]*m3M2[0][2];
  (*lpm3M)[0][1] = m3M1[0][0]*m3M2[1][0] + m3M1[0][1]*m3M2[1][1]
                 + m3M1[0][2]*m3M2[1][2];
  (*lpm3M)[0][2] = m3M1[0][0]*m3M2[2][0] + m3M1[0][1]*m3M2[2][1]
                 + m3M1[0][2]*m3M2[2][2];

  (*lpm3M)[1][0] = m3M1[1][0]*m3M2[0][0] + m3M1[1][1]*m3M2[0][1]
                 + m3M1[1][2]*m3M2[0][2];
  (*lpm3M)[1][1] = m3M1[1][0]*m3M2[1][0] + m3M1[1][1]*m3M2[1][1]
                 + m3M1[1][2]*m3M2[1][2];
  (*lpm3M)[1][2] = m3M1[1][0]*m3M2[2][0] + m3M1[1][1]*m3M2[2][1]
                 + m3M1[1][2]*m3M2[2][2];

  (*lpm3M)[2][0] = m3M1[2][0]*m3M2[0][0] + m3M1[2][1]*m3M2[0][1]
                 + m3M1[2][2]*m3M2[0][2];
  (*lpm3M)[2][1] = m3M1[2][0]*m3M2[1][0] + m3M1[2][1]*m3M2[1][1]
                 + m3M1[2][2]*m3M2[1][2];
  (*lpm3M)[2][2] = m3M1[2][0]*m3M2[2][0] + m3M1[2][1]*m3M2[2][1]
                 + m3M1[2][2]*m3M2[2][2];
  }

void vMultMxM(CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)
  {                                                                   // M=tM1*M2
  (*lpm3M)[0][0] = m3M1[0][0]*m3M2[0][0] + m3M1[1][0]*m3M2[1][0]
                 + m3M1[2][0]*m3M2[2][0];
  (*lpm3M)[0][1] = m3M1[0][0]*m3M2[0][1] + m3M1[1][0]*m3M2[1][1]
                 + m3M1[2][0]*m3M2[2][1];
  (*lpm3M)[0][2] = m3M1[0][0]*m3M2[0][2] + m3M1[1][0]*m3M2[1][2]
                 + m3M1[2][0]*m3M2[2][2];

  (*lpm3M)[1][0] = m3M1[0][1]*m3M2[0][0] + m3M1[1][1]*m3M2[1][0]
                 + m3M1[2][1]*m3M2[2][0];
  (*lpm3M)[1][1] = m3M1[0][1]*m3M2[0][1] + m3M1[1][1]*m3M2[1][1]
                 + m3M1[2][1]*m3M2[2][1];
  (*lpm3M)[1][2] = m3M1[0][1]*m3M2[0][2] + m3M1[1][1]*m3M2[1][2]
                 + m3M1[2][1]*m3M2[2][2];

  (*lpm3M)[2][0] = m3M1[0][2]*m3M2[0][0] + m3M1[1][2]*m3M2[1][0]
                 + m3M1[2][2]*m3M2[2][0];
  (*lpm3M)[2][1] = m3M1[0][2]*m3M2[0][1] + m3M1[1][2]*m3M2[1][1]
                 + m3M1[2][2]*m3M2[2][1];
  (*lpm3M)[2][2] = m3M1[0][2]*m3M2[0][2] + m3M1[1][2]*m3M2[1][2]
                 + m3M1[2][2]*m3M2[2][2];
  }

/**********************************************************************************************************************/
/*                                                4x4                                                                 */
/**********************************************************************************************************************/

void vAdd(CMat4x4* lpm4M, const CMat4x4& m4M1)                        // M+=M1
  {
  (*lpm4M)[0][0] += m4M1[0][0];
  (*lpm4M)[0][1] += m4M1[0][1];
  (*lpm4M)[0][2] += m4M1[0][2];
  (*lpm4M)[0][3] += m4M1[0][3];

  (*lpm4M)[1][0] += m4M1[1][0];
  (*lpm4M)[1][1] += m4M1[1][1];
  (*lpm4M)[1][2] += m4M1[1][2];
  (*lpm4M)[1][3] += m4M1[1][3];

  (*lpm4M)[2][0] += m4M1[2][0];
  (*lpm4M)[2][1] += m4M1[2][1];
  (*lpm4M)[2][2] += m4M1[2][2];
  (*lpm4M)[2][3] += m4M1[2][3];

  (*lpm4M)[3][0] += m4M1[3][0];
  (*lpm4M)[3][1] += m4M1[3][1];
  (*lpm4M)[3][2] += m4M1[3][2];
  (*lpm4M)[3][3] += m4M1[3][3];
  }

void vSub(CMat4x4* lpm4M, const CMat4x4& m4M1)                        // M-=M1
  {
  (*lpm4M)[0][0] -= m4M1[0][0];
  (*lpm4M)[0][1] -= m4M1[0][1];
  (*lpm4M)[0][2] -= m4M1[0][2];
  (*lpm4M)[0][3] -= m4M1[0][3];

  (*lpm4M)[1][0] -= m4M1[1][0];
  (*lpm4M)[1][1] -= m4M1[1][1];
  (*lpm4M)[1][2] -= m4M1[1][2];
  (*lpm4M)[1][3] -= m4M1[1][3];

  (*lpm4M)[2][0] -= m4M1[2][0];
  (*lpm4M)[2][1] -= m4M1[2][1];
  (*lpm4M)[2][2] -= m4M1[2][2];
  (*lpm4M)[2][3] -= m4M1[2][3];

  (*lpm4M)[3][0] -= m4M1[3][0];
  (*lpm4M)[3][1] -= m4M1[3][1];
  (*lpm4M)[3][2] -= m4M1[3][2];
  (*lpm4M)[3][3] -= m4M1[3][3];
  }

void vMul(CMat4x4* lpm4M, const FLOAT fK)                             // M*=k
  {
  (*lpm4M)[0][0] *= fK;
  (*lpm4M)[0][1] *= fK;
  (*lpm4M)[0][2] *= fK;
  (*lpm4M)[0][3] *= fK;

  (*lpm4M)[1][0] *= fK;
  (*lpm4M)[1][1] *= fK;
  (*lpm4M)[1][2] *= fK;
  (*lpm4M)[1][3] *= fK;

  (*lpm4M)[2][0] *= fK;
  (*lpm4M)[2][1] *= fK;
  (*lpm4M)[2][2] *= fK;
  (*lpm4M)[2][3] *= fK;

  (*lpm4M)[3][0] *= fK;
  (*lpm4M)[3][1] *= fK;
  (*lpm4M)[3][2] *= fK;
  (*lpm4M)[3][3] *= fK;
  }

void vMul(CMat4x4* lpm4M, const CMat4x4& m4M1)                        // M*=M1
  {
  *lpm4M = (*lpm4M)*m4M1;
  }

void vDiv(CMat4x4* lpm4M, const FLOAT fK)                             // M/=k
  {
  FLOAT fKinv = 1.f/fK;

  (*lpm4M)[0][0] *= fKinv;
  (*lpm4M)[0][1] *= fKinv;
  (*lpm4M)[0][2] *= fKinv;
  (*lpm4M)[0][3] *= fKinv;

  (*lpm4M)[1][0] *= fKinv;
  (*lpm4M)[1][1] *= fKinv;
  (*lpm4M)[1][2] *= fKinv;
  (*lpm4M)[1][3] *= fKinv;

  (*lpm4M)[2][0] *= fKinv;
  (*lpm4M)[2][1] *= fKinv;
  (*lpm4M)[2][2] *= fKinv;
  (*lpm4M)[2][3] *= fKinv;

  (*lpm4M)[3][0] *= fKinv;
  (*lpm4M)[3][1] *= fKinv;
  (*lpm4M)[3][2] *= fKinv;
  (*lpm4M)[3][3] *= fKinv;
  }

void vDiv(CMat4x4* lpm4M, const CMat4x4& m4M1)                        // M/=M1 (* inverse)
  {
  *lpm4M = (*lpm4M)*(m4M1.m4Inverse());
  }

void vNeg(CMat4x4* lpm4M, const CMat4x4& m4M1)                        // M=-M1
  {
  (*lpm4M)[0][0] = -m4M1[0][0];
  (*lpm4M)[0][1] = -m4M1[0][1];
  (*lpm4M)[0][2] = -m4M1[0][2];
  (*lpm4M)[0][3] = -m4M1[0][3];

  (*lpm4M)[1][0] = -m4M1[1][0];
  (*lpm4M)[1][1] = -m4M1[1][1];
  (*lpm4M)[1][2] = -m4M1[1][2];
  (*lpm4M)[1][3] = -m4M1[1][3];

  (*lpm4M)[2][0] = -m4M1[2][0];
  (*lpm4M)[2][1] = -m4M1[2][1];
  (*lpm4M)[2][2] = -m4M1[2][2];
  (*lpm4M)[2][3] = -m4M1[2][3];

  (*lpm4M)[3][0] = -m4M1[3][0];
  (*lpm4M)[3][1] = -m4M1[3][1];
  (*lpm4M)[3][2] = -m4M1[3][2];
  (*lpm4M)[3][3] = -m4M1[3][3];
  }

void vMul(CMat4x4* lpm4M, const CMat4x4& m4M1,const FLOAT fK)         // M=M1*k
  {
  (*lpm4M)[0][0] = m4M1[0][0]*fK;
  (*lpm4M)[0][1] = m4M1[0][1]*fK;
  (*lpm4M)[0][2] = m4M1[0][2]*fK;
  (*lpm4M)[0][3] = m4M1[0][3]*fK;

  (*lpm4M)[1][0] = m4M1[1][0]*fK;
  (*lpm4M)[1][1] = m4M1[1][1]*fK;
  (*lpm4M)[1][2] = m4M1[1][2]*fK;
  (*lpm4M)[1][3] = m4M1[1][3]*fK;

  (*lpm4M)[2][0] = m4M1[2][0]*fK;
  (*lpm4M)[2][1] = m4M1[2][1]*fK;
  (*lpm4M)[2][2] = m4M1[2][2]*fK;
  (*lpm4M)[2][3] = m4M1[2][3]*fK;

  (*lpm4M)[3][0] = m4M1[3][0]*fK;
  (*lpm4M)[3][1] = m4M1[3][1]*fK;
  (*lpm4M)[3][2] = m4M1[3][2]*fK;
  (*lpm4M)[3][3] = m4M1[3][3]*fK;
  }

void vDiv(CMat4x4* lpm4M, const CMat4x4& m4M1,const FLOAT fK)         // M=M1/k
  {
  FLOAT fKinv = 1.f/fK;

  (*lpm4M)[0][0] = m4M1[0][0]*fKinv;
  (*lpm4M)[0][1] = m4M1[0][1]*fKinv;
  (*lpm4M)[0][2] = m4M1[0][2]*fKinv;
  (*lpm4M)[0][3] = m4M1[0][3]*fKinv;

  (*lpm4M)[1][0] = m4M1[1][0]*fKinv;
  (*lpm4M)[1][1] = m4M1[1][1]*fKinv;
  (*lpm4M)[1][2] = m4M1[1][2]*fKinv;
  (*lpm4M)[1][3] = m4M1[1][3]*fKinv;

  (*lpm4M)[2][0] = m4M1[2][0]*fKinv;
  (*lpm4M)[2][1] = m4M1[2][1]*fKinv;
  (*lpm4M)[2][2] = m4M1[2][2]*fKinv;
  (*lpm4M)[2][3] = m4M1[2][3]*fKinv;

  (*lpm4M)[3][0] = m4M1[3][0]*fKinv;
  (*lpm4M)[3][1] = m4M1[3][1]*fKinv;
  (*lpm4M)[3][2] = m4M1[3][2]*fKinv;
  (*lpm4M)[3][3] = m4M1[3][3]*fKinv;
  }

void vSub(CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)    // M=M1-M2
  {
  (*lpm4M)[0][0] = m4M1[0][0] - m4M2[0][0];
  (*lpm4M)[0][1] = m4M1[0][1] - m4M2[0][1];
  (*lpm4M)[0][2] = m4M1[0][2] - m4M2[0][2];
  (*lpm4M)[0][3] = m4M1[0][3] - m4M2[0][3];

  (*lpm4M)[1][0] = m4M1[1][0] - m4M2[1][0];
  (*lpm4M)[1][1] = m4M1[1][1] - m4M2[1][1];
  (*lpm4M)[1][2] = m4M1[1][2] - m4M2[1][2];
  (*lpm4M)[1][3] = m4M1[1][3] - m4M2[1][3];

  (*lpm4M)[2][0] = m4M1[2][0] - m4M2[2][0];
  (*lpm4M)[2][1] = m4M1[2][1] - m4M2[2][1];
  (*lpm4M)[2][2] = m4M1[2][2] - m4M2[2][2];
  (*lpm4M)[2][3] = m4M1[2][3] - m4M2[2][3];

  (*lpm4M)[3][0] = m4M1[3][0] - m4M2[3][0];
  (*lpm4M)[3][1] = m4M1[3][1] - m4M2[3][1];
  (*lpm4M)[3][2] = m4M1[3][2] - m4M2[3][2];
  (*lpm4M)[3][3] = m4M1[3][3] - m4M2[3][3];
  }

void vAdd(CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)    // M=M1+M2
  {
  (*lpm4M)[0][0] = m4M1[0][0] + m4M2[0][0];
  (*lpm4M)[0][1] = m4M1[0][1] + m4M2[0][1];
  (*lpm4M)[0][2] = m4M1[0][2] + m4M2[0][2];
  (*lpm4M)[0][3] = m4M1[0][3] + m4M2[0][3];

  (*lpm4M)[1][0] = m4M1[1][0] + m4M2[1][0];
  (*lpm4M)[1][1] = m4M1[1][1] + m4M2[1][1];
  (*lpm4M)[1][2] = m4M1[1][2] + m4M2[1][2];
  (*lpm4M)[1][3] = m4M1[1][3] + m4M2[1][3];

  (*lpm4M)[2][0] = m4M1[2][0] + m4M2[2][0];
  (*lpm4M)[2][1] = m4M1[2][1] + m4M2[2][1];
  (*lpm4M)[2][2] = m4M1[2][2] + m4M2[2][2];
  (*lpm4M)[2][3] = m4M1[2][3] + m4M2[2][3];

  (*lpm4M)[3][0] = m4M1[3][0] + m4M2[3][0];
  (*lpm4M)[3][1] = m4M1[3][1] + m4M2[3][1];
  (*lpm4M)[3][2] = m4M1[3][2] + m4M2[3][2];
  (*lpm4M)[3][3] = m4M1[3][3] + m4M2[3][3];
  }

void vMul(CMat4x4* lpm4M,  const CMat4x4& m4M1, const CMat4x4& m4M2)  // M=M1*M2
  {
  (*lpm4M)[0][0] = m4M1[0][0]*m4M2[0][0] + m4M1[0][1]*m4M2[1][0]
                 + m4M1[0][2]*m4M2[2][0] + m4M1[0][3]*m4M2[3][0];
  (*lpm4M)[0][1] = m4M1[0][0]*m4M2[0][1] + m4M1[0][1]*m4M2[1][1]
                 + m4M1[0][2]*m4M2[2][1] + m4M1[0][3]*m4M2[3][1];
  (*lpm4M)[0][2] = m4M1[0][0]*m4M2[0][2] + m4M1[0][1]*m4M2[1][2]
                 + m4M1[0][2]*m4M2[2][2] + m4M1[0][3]*m4M2[3][2];
  (*lpm4M)[0][3] = m4M1[0][0]*m4M2[0][3] + m4M1[0][1]*m4M2[1][3]
                 + m4M1[0][2]*m4M2[2][3] + m4M1[0][3]*m4M2[3][3];

  (*lpm4M)[1][0] = m4M1[1][0]*m4M2[0][0] + m4M1[1][1]*m4M2[1][0]
                 + m4M1[1][2]*m4M2[2][0] + m4M1[1][3]*m4M2[3][0];
  (*lpm4M)[1][1] = m4M1[1][0]*m4M2[0][1] + m4M1[1][1]*m4M2[1][1]
                 + m4M1[1][2]*m4M2[2][1] + m4M1[1][3]*m4M2[3][1];
  (*lpm4M)[1][2] = m4M1[1][0]*m4M2[0][2] + m4M1[1][1]*m4M2[1][2]
                 + m4M1[1][2]*m4M2[2][2] + m4M1[1][3]*m4M2[3][2];
  (*lpm4M)[1][3] = m4M1[1][0]*m4M2[0][3] + m4M1[1][1]*m4M2[1][3]
                 + m4M1[1][2]*m4M2[2][3] + m4M1[1][3]*m4M2[3][3];

  (*lpm4M)[2][0] = m4M1[2][0]*m4M2[0][0] + m4M1[2][1]*m4M2[1][0]
                 + m4M1[2][2]*m4M2[2][0] + m4M1[2][3]*m4M2[3][0];
  (*lpm4M)[2][1] = m4M1[2][0]*m4M2[0][1] + m4M1[2][1]*m4M2[1][1]
                 + m4M1[2][2]*m4M2[2][1] + m4M1[2][3]*m4M2[3][1];
  (*lpm4M)[2][2] = m4M1[2][0]*m4M2[0][2] + m4M1[2][1]*m4M2[1][2]
                 + m4M1[2][2]*m4M2[2][2] + m4M1[2][3]*m4M2[3][2];
  (*lpm4M)[2][3] = m4M1[2][0]*m4M2[0][3] + m4M1[2][1]*m4M2[1][3]
                 + m4M1[2][2]*m4M2[2][3] + m4M1[2][3]*m4M2[3][3];

  (*lpm4M)[3][0] = m4M1[3][0]*m4M2[0][0] + m4M1[3][1]*m4M2[1][0]
                 + m4M1[3][2]*m4M2[2][0] + m4M1[3][3]*m4M2[3][0];
  (*lpm4M)[3][1] = m4M1[3][0]*m4M2[0][1] + m4M1[3][1]*m4M2[1][1]
                 + m4M1[3][2]*m4M2[2][1] + m4M1[3][3]*m4M2[3][1];
  (*lpm4M)[3][2] = m4M1[3][0]*m4M2[0][2] + m4M1[3][1]*m4M2[1][2]
                 + m4M1[3][2]*m4M2[2][2] + m4M1[3][3]*m4M2[3][2];
  (*lpm4M)[3][3] = m4M1[3][0]*m4M2[0][3] + m4M1[3][1]*m4M2[1][3]
                 + m4M1[3][2]*m4M2[2][3] + m4M1[3][3]*m4M2[3][3];
  }

void vMulTerm(CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)
  {                                                                   // M=M1*M2 (term by term)
  (*lpm4M)[0][0] = m4M1[0][0] * m4M2[0][0];
  (*lpm4M)[0][1] = m4M1[0][1] * m4M2[0][1];
  (*lpm4M)[0][2] = m4M1[0][2] * m4M2[0][2];
  (*lpm4M)[0][3] = m4M1[0][3] * m4M2[0][3];

  (*lpm4M)[1][0] = m4M1[1][0] * m4M2[1][0];
  (*lpm4M)[1][1] = m4M1[1][1] * m4M2[1][1];
  (*lpm4M)[1][2] = m4M1[1][2] * m4M2[1][2];
  (*lpm4M)[1][3] = m4M1[1][3] * m4M2[1][3];

  (*lpm4M)[2][0] = m4M1[2][0] * m4M2[2][0];
  (*lpm4M)[2][1] = m4M1[2][1] * m4M2[2][1];
  (*lpm4M)[2][2] = m4M1[2][2] * m4M2[2][2];
  (*lpm4M)[2][3] = m4M1[2][3] * m4M2[2][3];

  (*lpm4M)[3][0] = m4M1[3][0] * m4M2[3][0];
  (*lpm4M)[3][1] = m4M1[3][1] * m4M2[3][1];
  (*lpm4M)[3][2] = m4M1[3][2] * m4M2[3][2];
  (*lpm4M)[3][3] = m4M1[3][3] * m4M2[3][3];
  }

void vDiv(CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)    // M=M1/M2 (term by term)
  {
  (*lpm4M)[0][0] = m4M1[0][0] / m4M2[0][0];
  (*lpm4M)[0][1] = m4M1[0][1] / m4M2[0][1];
  (*lpm4M)[0][2] = m4M1[0][2] / m4M2[0][2];
  (*lpm4M)[0][3] = m4M1[0][3] / m4M2[0][3];

  (*lpm4M)[1][0] = m4M1[1][0] / m4M2[1][0];
  (*lpm4M)[1][1] = m4M1[1][1] / m4M2[1][1];
  (*lpm4M)[1][2] = m4M1[1][2] / m4M2[1][2];
  (*lpm4M)[1][3] = m4M1[1][3] / m4M2[1][3];

  (*lpm4M)[2][0] = m4M1[2][0] / m4M2[2][0];
  (*lpm4M)[2][1] = m4M1[2][1] / m4M2[2][1];
  (*lpm4M)[2][2] = m4M1[2][2] / m4M2[2][2];
  (*lpm4M)[2][3] = m4M1[2][3] / m4M2[2][3];

  (*lpm4M)[3][0] = m4M1[3][0] / m4M2[3][0];
  (*lpm4M)[3][1] = m4M1[3][1] / m4M2[3][1];
  (*lpm4M)[3][2] = m4M1[3][2] / m4M2[3][2];
  (*lpm4M)[3][3] = m4M1[3][3] / m4M2[3][3];
  }

void vTranspose(CMat4x4* lpm4M, const CMat4x4& m4M1)                  // M=tM1
  {
  (*lpm4M)[0][0] = m4M1[0][0];
  (*lpm4M)[0][1] = m4M1[1][0];
  (*lpm4M)[0][2] = m4M1[2][0];
  (*lpm4M)[0][3] = m4M1[3][0];

  (*lpm4M)[1][0] = m4M1[0][1];
  (*lpm4M)[1][1] = m4M1[1][1];
  (*lpm4M)[1][2] = m4M1[2][1];
  (*lpm4M)[1][3] = m4M1[3][1];

  (*lpm4M)[2][0] = m4M1[0][2];
  (*lpm4M)[2][1] = m4M1[1][2];
  (*lpm4M)[2][2] = m4M1[2][2];
  (*lpm4M)[2][3] = m4M1[3][2];

  (*lpm4M)[3][0] = m4M1[0][3];
  (*lpm4M)[3][1] = m4M1[1][3];
  (*lpm4M)[3][2] = m4M1[2][3];
  (*lpm4M)[3][3] = m4M1[3][3];
  }

void vMulMxtM(CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)
  {                                                                   // M=M1*tM2
  (*lpm4M)[0][0] = m4M1[0][0]*m4M2[0][0] + m4M1[0][1]*m4M2[0][1]
                 + m4M1[0][2]*m4M2[0][2] + m4M1[0][3]*m4M2[0][3];
  (*lpm4M)[0][1] = m4M1[0][0]*m4M2[1][0] + m4M1[0][1]*m4M2[1][1]
                 + m4M1[0][2]*m4M2[1][2] + m4M1[0][3]*m4M2[1][3];
  (*lpm4M)[0][2] = m4M1[0][0]*m4M2[2][0] + m4M1[0][1]*m4M2[2][1]
                 + m4M1[0][2]*m4M2[2][2] + m4M1[0][3]*m4M2[2][3];
  (*lpm4M)[0][3] = m4M1[0][0]*m4M2[3][0] + m4M1[0][1]*m4M2[3][1]
                 + m4M1[0][2]*m4M2[3][2] + m4M1[0][3]*m4M2[3][3];

  (*lpm4M)[1][0] = m4M1[1][0]*m4M2[0][0] + m4M1[1][1]*m4M2[0][1]
                 + m4M1[1][2]*m4M2[0][2] + m4M1[1][3]*m4M2[0][3];
  (*lpm4M)[1][1] = m4M1[1][0]*m4M2[1][0] + m4M1[1][1]*m4M2[1][1]
                 + m4M1[1][2]*m4M2[1][2] + m4M1[1][3]*m4M2[1][3];
  (*lpm4M)[1][2] = m4M1[1][0]*m4M2[2][0] + m4M1[1][1]*m4M2[2][1]
                 + m4M1[1][2]*m4M2[2][2] + m4M1[1][3]*m4M2[2][3];
  (*lpm4M)[1][3] = m4M1[1][0]*m4M2[3][0] + m4M1[1][1]*m4M2[3][1]
                 + m4M1[1][2]*m4M2[3][2] + m4M1[1][3]*m4M2[3][3];

  (*lpm4M)[2][0] = m4M1[2][0]*m4M2[0][0] + m4M1[2][1]*m4M2[0][1]
                 + m4M1[2][2]*m4M2[0][2] + m4M1[2][3]*m4M2[0][3];
  (*lpm4M)[2][1] = m4M1[2][0]*m4M2[1][0] + m4M1[2][1]*m4M2[1][1]
                 + m4M1[2][2]*m4M2[1][2] + m4M1[2][3]*m4M2[1][3];
  (*lpm4M)[2][2] = m4M1[2][0]*m4M2[2][0] + m4M1[2][1]*m4M2[2][1]
                 + m4M1[2][2]*m4M2[2][2] + m4M1[2][3]*m4M2[2][3];
  (*lpm4M)[2][3] = m4M1[2][0]*m4M2[3][0] + m4M1[2][1]*m4M2[3][1]
                 + m4M1[2][2]*m4M2[3][2] + m4M1[2][3]*m4M2[3][3];

  (*lpm4M)[3][0] = m4M1[3][0]*m4M2[0][0] + m4M1[3][1]*m4M2[0][1]
                 + m4M1[3][2]*m4M2[0][2] + m4M1[3][3]*m4M2[0][3];
  (*lpm4M)[3][1] = m4M1[3][0]*m4M2[1][0] + m4M1[3][1]*m4M2[1][1]
                 + m4M1[3][2]*m4M2[1][2] + m4M1[3][3]*m4M2[1][3];
  (*lpm4M)[3][2] = m4M1[3][0]*m4M2[2][0] + m4M1[3][1]*m4M2[2][1]
                 + m4M1[3][2]*m4M2[2][2] + m4M1[3][3]*m4M2[2][3];
  (*lpm4M)[3][3] = m4M1[3][0]*m4M2[3][0] + m4M1[3][1]*m4M2[3][1]
                 + m4M1[3][2]*m4M2[3][2] + m4M1[3][3]*m4M2[3][3];
  }

void vMultMxM(CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)
  {                                                                   // M=tM1*M2
  (*lpm4M)[0][0] = m4M1[0][0]*m4M2[0][0] + m4M1[1][0]*m4M2[1][0]
                 + m4M1[2][0]*m4M2[2][0] + m4M1[3][0]*m4M2[3][0];
  (*lpm4M)[0][1] = m4M1[0][0]*m4M2[0][1] + m4M1[1][0]*m4M2[1][1]
                 + m4M1[2][0]*m4M2[2][1] + m4M1[3][0]*m4M2[3][1];
  (*lpm4M)[0][2] = m4M1[0][0]*m4M2[0][2] + m4M1[1][0]*m4M2[1][2]
                 + m4M1[2][0]*m4M2[2][2] + m4M1[3][0]*m4M2[3][2];
  (*lpm4M)[0][3] = m4M1[0][0]*m4M2[0][3] + m4M1[1][0]*m4M2[1][3]
                 + m4M1[2][0]*m4M2[2][3] + m4M1[3][0]*m4M2[3][3];

  (*lpm4M)[1][0] = m4M1[0][1]*m4M2[0][0] + m4M1[1][1]*m4M2[1][0]
                 + m4M1[2][1]*m4M2[2][0] + m4M1[3][1]*m4M2[3][0];
  (*lpm4M)[1][1] = m4M1[0][1]*m4M2[0][1] + m4M1[1][1]*m4M2[1][1]
                 + m4M1[2][1]*m4M2[2][1] + m4M1[3][1]*m4M2[3][1];
  (*lpm4M)[1][2] = m4M1[0][1]*m4M2[0][2] + m4M1[1][1]*m4M2[1][2]
                 + m4M1[2][1]*m4M2[2][2] + m4M1[3][1]*m4M2[3][2];
  (*lpm4M)[1][3] = m4M1[0][1]*m4M2[0][3] + m4M1[1][1]*m4M2[1][3]
                 + m4M1[2][1]*m4M2[2][3] + m4M1[3][1]*m4M2[3][3];

  (*lpm4M)[2][0] = m4M1[0][2]*m4M2[0][0] + m4M1[1][2]*m4M2[1][0]
                 + m4M1[2][2]*m4M2[2][0] + m4M1[3][2]*m4M2[3][0];
  (*lpm4M)[2][1] = m4M1[0][2]*m4M2[0][1] + m4M1[1][2]*m4M2[1][1]
                 + m4M1[2][2]*m4M2[2][1] + m4M1[3][2]*m4M2[3][1];
  (*lpm4M)[2][2] = m4M1[0][2]*m4M2[0][2] + m4M1[1][2]*m4M2[1][2]
                 + m4M1[2][2]*m4M2[2][2] + m4M1[3][2]*m4M2[3][2];
  (*lpm4M)[2][3] = m4M1[0][2]*m4M2[0][3] + m4M1[1][2]*m4M2[1][3]
                 + m4M1[2][2]*m4M2[2][3] + m4M1[3][2]*m4M2[3][3];

  (*lpm4M)[3][0] = m4M1[0][3]*m4M2[0][0] + m4M1[1][3]*m4M2[1][0]
                 + m4M1[2][3]*m4M2[2][0] + m4M1[3][3]*m4M2[3][0];
  (*lpm4M)[3][1] = m4M1[0][3]*m4M2[0][1] + m4M1[1][3]*m4M2[1][1]
                 + m4M1[2][3]*m4M2[2][1] + m4M1[3][3]*m4M2[3][1];
  (*lpm4M)[3][2] = m4M1[0][3]*m4M2[0][2] + m4M1[1][3]*m4M2[1][2]
                 + m4M1[2][3]*m4M2[2][2] + m4M1[3][3]*m4M2[3][2];
  (*lpm4M)[3][3] = m4M1[0][3]*m4M2[0][3] + m4M1[1][3]*m4M2[1][3]
                 + m4M1[2][3]*m4M2[2][3] + m4M1[3][3]*m4M2[3][3];
  }
