/**********************************************************************************************************************/
/*                                                QUATERNION CLASS                                                    */
/* 26/08/00, M : creation of the file                                                                                 */
/**********************************************************************************************************************/

#ifndef   _QUATERNION_H_
#define   _QUATERNION_H_

/************************************************ INCLUDE *************************************************************/

#include  "CVectMat.h"                                      // vectors & matrices

/************************************************ DEFINE VALUES *******************************************************/

/************************************************ MACROS **************************************************************/

/************************************************ CLASSES *************************************************************/

/**********************************************************************************************************************/
/*                                                QUATERNION                                                          */
/**********************************************************************************************************************/

class     CQuaternion
  {
// public methods

  public:

    // constructors

    inline          CQuaternion         (void);
    inline          CQuaternion         (const CVect3D&     v3Vector,const FLOAT fScalar);
    inline          CQuaternion         (const CVect4D&     v4Quat);                      // x/y/z=vector, w=scalar
    inline          CQuaternion         (const FLOAT        fX,const FLOAT fY,const FLOAT fZ,const FLOAT fW);
    inline          CQuaternion         (const CQuaternion& quatQ);

    // assignment operators

    CQuaternion&           operator   = (const CQuaternion& quatQ);                       // assignment
    CQuaternion&           operator  += (const CQuaternion& quatQ);                       // incrementation
    CQuaternion&           operator  -= (const CQuaternion& quatQ);                       // decrementation
    CQuaternion&           operator  *= (const CQuaternion& quatQ);                       // multiplication
    CQuaternion&           operator  *= (const FLOAT        fK);                          // multiplication by a constant
    CQuaternion&           operator  /= (const FLOAT        fK);                          // division       by a constant

    // friends operators

    friend CQuaternion inline operator -(const CQuaternion& quatQ);                                 // -Q
    friend CQuaternion inline operator *(const FLOAT        fK,    const CQuaternion& quatQ);       // k*Q 
    friend CQuaternion inline operator *(const CQuaternion& quatQ, const FLOAT        fK);          // Q*k
    friend CQuaternion inline operator /(const CQuaternion& quatQ, const FLOAT        fK);          // Q/k
    friend CQuaternion inline operator +(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1+Q2
    friend CQuaternion inline operator -(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1-Q2
    friend CQuaternion inline operator *(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1*Q2
    friend bool        inline operator==(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1==Q2
    friend bool        inline operator!=(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1!=Q2

    friend ostream&        operator  << (ostream& s,      CQuaternion& quatQ);            // output
    friend ostream&        operator  << (ostream& s,const CQuaternion& quatQ);            // output

    // special functions

    friend CQuaternion     quatSlerp    (const CQuaternion& quatQ1,const CQuaternion& quatQ2,const FLOAT fT);
    CQuaternion            quatConjugate(void) const;
    FLOAT  inline          fLength      (void) const;
    FLOAT  inline          fLength2     (void) const;

    // conversions

    void            vFromAxisAngle      (const CVect3D&     v3Axis,const FLOAT        fAngle);
    void            vToAxisAngle        (      CVect3D*   lpv3Axis,      FLOAT*     lpfAngle);

    void            vFromRotation       (const CMat3x3&     m3M);
    void            vFromRotation       (const CMat4x4&     m4M);
    void            vToRotation         (      CMat3x3*   lpm3M) const;
    void            vToRotation         (      CMat4x4*   lpm4M) const;

    // get/set

    inline FLOAT    fGetScalar          (void);
    inline CVect3D& v3GetVector         (void);

    inline void     vSetScalar          (const FLOAT        fScalar);
    inline void     vSetVector          (const CVect3D&     v3Vector);
    inline void     vSetVector          (const FLOAT        fX,const FLOAT fY,const FLOAT fZ);
    inline void     vSet                (const FLOAT        fX,const FLOAT fY,const FLOAT fZ,const FLOAT fW);
    inline void     vSet                (const CVect4D&     v4Quat);
    inline void     vSet                (const CQuaternion& quatQ);

    // friend functions

    friend void     vAdd                (CQuaternion* lpquatQ, const CQuaternion& quatQ1);                              // Q+=Q1
    friend void     vSub                (CQuaternion* lpquatQ, const CQuaternion& quatQ1);                              // Q-=Q1
    friend void     vMul                (CQuaternion* lpquatQ, const CQuaternion& quatQ1);                              // Q*=Q1
    friend void     vMul                (CQuaternion* lpquatQ, const FLOAT        fK);                                  // Q*=k
    friend void     vDiv                (CQuaternion* lpquatQ, const FLOAT        fK);                                  // Q/=k

    friend void     vNeg                (CQuaternion* lpquatQ, const CQuaternion& quatQ1);                              // Q=-Q1
    friend void     vMul                (CQuaternion* lpquatQ, const CQuaternion& quatQ1,const FLOAT        fK);        // Q=Q1*k
    friend void     vDiv                (CQuaternion* lpquatQ, const CQuaternion& quatQ1,const FLOAT        fK);        // Q=Q1/k
    friend void     vAdd                (CQuaternion* lpquatQ, const CQuaternion& quatQ1,const CQuaternion& quatQ2);    // Q=Q1+Q2
    friend void     vSub                (CQuaternion* lpquatQ, const CQuaternion& quatQ1,const CQuaternion& quatQ2);    // Q=Q1-Q2

    friend void     vMul                (CQuaternion* lpquatQ, const CQuaternion& quatQ1,const CQuaternion& quatQ2);    // Q=Q1*Q2

// protected members

  protected:
                                                                                          // ! MUST be the same size as a CVect4D
    CVect3D         m_v3Vector;                                                           // imaginary part
    FLOAT           m_fScalar;                                                            // real      part
  };

/************************************************ INLINE **************************************************************/

//#include  "CQuaternion.inl"

/**********************************************************************************************************************/

#endif
